@@ -52,6 +52,11 @@
 #include <utility>
 #include <thread>
 
+#if defined(__linux__)
+#include <xf86drm.h>
+#include <amdgpu.h>
+#endif
+
 #include "core/inc/hsa_ext_interface.h"
 #include "core/inc/hsa_internal.h"
 #include "core/inc/hsa_ext_amd_impl.h"
@@ -70,6 +75,12 @@
 #include "core/inc/amd_loader_context.hpp"
 #include "core/inc/amd_hsa_code.hpp"
 
+#if defined(__clang__)
+#if __has_feature(address_sanitizer)
+#define SANITIZER_AMDGPU 1
+#endif
+#endif
+
 //---------------------------------------------------------------------------//
 //    Constants                                                              //
 //---------------------------------------------------------------------------//
@@ -86,6 +97,7 @@ namespace AMD {
 
 namespace core {
 extern bool g_use_interrupt_wait;
+extern bool g_use_mwaitx;
 
 /// @brief  Runtime class provides the following functions:
 /// - open and close connection to kernel driver.
@@ -111,6 +123,7 @@ class Runtime {
   struct KfdVersion_t {
     HsaVersionInfo version;
     bool supports_exception_debugging;
+    bool supports_event_age;
   };
 
   /// @brief Open connection to kernel driver and increment reference count.
@@ -123,6 +136,9 @@ class Runtime {
   /// @retval True if the connection to kernel driver is opened.
   static bool IsOpen();
 
+  // @brief Callback handler for HW Exceptions.
+  static bool HwExceptionHandler(hsa_signal_value_t val, void* arg);
+
   // @brief Callback handler for VM fault access.
   static bool VMFaultHandler(hsa_signal_value_t val, void* arg);
 
@@ -227,6 +243,35 @@ class Runtime {
                           core::Agent* src_agent, size_t size,
                           std::vector<core::Signal*>& dep_signals, core::Signal& completion_signal);
 
+  /// @brief Non-blocking memory copy from src to dst on engine_id.
+  ///
+  /// @details All semantics and params are dentical to CopyMemory
+  ///  with the exception of engine_id.
+  ///
+  /// @param [in] engine_id Target engine to copy on.
+  ///
+  /// @param [in] force_copy_on_sdma By default, a blit kernel copy is used
+  /// when dst_agent == src_agent.  Setting this to true will force the copy
+  /// over SDMA1.
+  ///
+  /// @retval ::HSA_STATUS_SUCCESS if copy command has been submitted
+  /// successfully to the agent DMA queue.
+  hsa_status_t CopyMemoryOnEngine(void* dst, core::Agent* dst_agent, const void* src,
+                          core::Agent* src_agent, size_t size,
+                          std::vector<core::Signal*>& dep_signals, core::Signal& completion_signal,
+                          hsa_amd_sdma_engine_id_t  engine_id, bool force_copy_on_sdma);
+
+  /// @brief Return SDMA availability status for copy direction
+  ///
+  /// @param [in] dst_agent Destination agent.
+  /// @param [in] src_agent Source agent.
+  /// @param [out] engine_ids_mask Mask of engine_ids.
+  ///
+  /// @retval HSA_STATUS_SUCCESS DMA engines are available
+  /// @retval HSA_STATUS_ERROR_OUT_OF_RESOURCES DMA engines are not available
+  hsa_status_t CopyMemoryStatus(core::Agent* dst_agent, core::Agent* src_agent,
+                                uint32_t *engine_ids_mask);
+
   /// @brief Fill the first @p count of uint32_t in ptr with value.
   ///
   /// @param [in] ptr Memory address to be filled.
@@ -308,6 +353,44 @@ class Runtime {
   hsa_status_t SvmPrefetch(void* ptr, size_t size, hsa_agent_t agent, uint32_t num_dep_signals,
                            const hsa_signal_t* dep_signals, hsa_signal_t completion_signal);
 
+  hsa_status_t DmaBufExport(const void* ptr, size_t size, int* dmabuf, uint64_t* offset);
+
+  hsa_status_t DmaBufClose(int dmabuf);
+
+  hsa_status_t VMemoryAddressReserve(void** ptr, size_t size, uint64_t address, uint64_t flags);
+
+  hsa_status_t VMemoryAddressFree(void* ptr, size_t size);
+
+  hsa_status_t VMemoryHandleCreate(const MemoryRegion* region, size_t size,
+                                   MemoryRegion::AllocateFlags alloc_flags,
+                                   uint64_t flags, hsa_amd_vmem_alloc_handle_t* memoryHandle);
+
+  hsa_status_t VMemoryHandleRelease(hsa_amd_vmem_alloc_handle_t memoryHandle);
+
+  hsa_status_t VMemoryHandleMap(void* va, size_t size, size_t in_offset,
+                                hsa_amd_vmem_alloc_handle_t memoryHandle, uint64_t flags);
+
+  hsa_status_t VMemoryHandleUnmap(void* va, size_t size);
+
+  hsa_status_t VMemorySetAccess(void* va, size_t size, const hsa_amd_memory_access_desc_t* desc,
+                                size_t desc_cnt);
+
+  hsa_status_t VMemoryGetAccess(const void* va, hsa_access_permission_t* perms,
+                                hsa_agent_t agent_handle);
+
+  hsa_status_t VMemoryExportShareableHandle(int* dmabuf_fd,
+                                            const hsa_amd_vmem_alloc_handle_t handle,
+                                            const uint64_t flags);
+
+  hsa_status_t VMemoryImportShareableHandle(const int dmabuf_fd,
+                                            hsa_amd_vmem_alloc_handle_t* handle);
+
+  hsa_status_t VMemoryRetainAllocHandle(hsa_amd_vmem_alloc_handle_t* memoryHandle, void* addr);
+
+  hsa_status_t VMemoryGetAllocPropertiesFromHandle(const hsa_amd_vmem_alloc_handle_t memoryHandle,
+                                                   const core::MemoryRegion** mem_region,
+                                                   hsa_amd_memory_type_t* type);
+
   const std::vector<Agent*>& cpu_agents() { return cpu_agents_; }
 
   const std::vector<Agent*>& gpu_agents() { return gpu_agents_; }
@@ -361,7 +444,12 @@ class Runtime {
 
   uint64_t sys_clock_freq() const { return sys_clock_freq_; }
 
-  void KfdVersion(const HsaVersionInfo& version) { kfd_version.version = version; }
+  void KfdVersion(const HsaVersionInfo& version) {
+    kfd_version.version = version;
+    if (version.KernelInterfaceMajorVersion == 1 &&
+      version.KernelInterfaceMinorVersion >= 14)
+      kfd_version.supports_event_age = true;
+  }
 
   void KfdVersion(bool exception_debugging) {
     kfd_version.supports_exception_debugging = exception_debugging;
@@ -369,13 +457,27 @@ class Runtime {
 
   KfdVersion_t KfdVersion() const { return kfd_version; }
 
+  bool VirtualMemApiSupported() const { return virtual_mem_api_supported_; }
+  bool XnackEnabled() const { return xnack_enabled_; }
+  void XnackEnabled(bool enable) { xnack_enabled_ = enable; }
+
  protected:
   static void AsyncEventsLoop(void*);
 
   struct AllocationRegion {
-    AllocationRegion() : region(NULL), size(0), user_ptr(nullptr) {}
-    AllocationRegion(const MemoryRegion* region_arg, size_t size_arg)
-        : region(region_arg), size(size_arg), user_ptr(nullptr) {}
+    AllocationRegion()
+        : region(NULL),
+          size(0),
+          size_requested(0),
+          alloc_flags(core::MemoryRegion::AllocateNoFlags),
+          user_ptr(nullptr) {}
+    AllocationRegion(const MemoryRegion* region_arg, size_t size_arg, size_t size_requested,
+                     MemoryRegion::AllocateFlags alloc_flags)
+        : region(region_arg),
+          size(size_arg),
+          size_requested(size_requested),
+          alloc_flags(alloc_flags),
+          user_ptr(nullptr) {}
 
     struct notifier_t {
       void* ptr;
@@ -384,7 +486,9 @@ class Runtime {
     };
 
     const MemoryRegion* region;
-    size_t size;
+    size_t size;           /* actual size = align_up(size_requested, granularity) */
+    size_t size_requested; /* size requested by user */
+    MemoryRegion::AllocateFlags alloc_flags;
     void* user_ptr;
     std::unique_ptr<std::vector<notifier_t>> notifiers;
   };
@@ -476,8 +580,8 @@ class Runtime {
   /// @brief Close tool libraries.
   void CloseTools();
 
-  // @brief Binds virtual memory access fault handler to this node.
-  void BindVmFaultHandler();
+  // @brief Binds Error handlers to this node.
+  void BindErrorHandlers();
 
   // @brief Acquire snapshot of system event handlers.
   // Returns a copy to avoid holding a lock during callbacks.
@@ -575,6 +679,12 @@ class Runtime {
   // @brief HSA signal to contain the VM fault event.
   Signal* vm_fault_signal_;
 
+  // @brief AMD HSA event to monitor for HW exceptions.
+  HsaEvent* hw_exception_event_;
+
+  // @brief HSA signal to contain the HW exceptionevent.
+  Signal* hw_exception_signal_;
+
   // Custom system event handlers.
   std::vector<std::pair<AMD::callback_t<hsa_amd_system_event_callback_t>, void*>>
       system_event_handlers_;
@@ -604,6 +714,113 @@ class Runtime {
 
   std::unique_ptr<AMD::SvmProfileControl> svm_profile_;
 
+ private:
+  void CheckVirtualMemApiSupport();
+  int GetAmdgpuDeviceArgs(Agent* agent, amdgpu_bo_handle bo, int* drm_fd, uint64_t* cpu_addr);
+
+  bool virtual_mem_api_supported_;
+  bool xnack_enabled_;
+
+  typedef void* ThunkHandle;
+
+  struct AddressHandle {
+    AddressHandle() : size(0), use_count(0) {}
+    AddressHandle(size_t size) : size(size), use_count(0) {}
+
+    size_t size;
+    int use_count;
+  };
+  std::map<const void*, AddressHandle> reserved_address_map_;  // Indexed by VA
+
+  struct MemoryHandle {
+    MemoryHandle() : region(NULL), size(0), ref_count(0), thunk_handle(NULL), alloc_flag(0) {}
+    MemoryHandle(const MemoryRegion* region, size_t size, uint64_t flags_unused,
+                 ThunkHandle thunk_handle, MemoryRegion::AllocateFlags alloc_flag)
+        : region(region),
+          size(size),
+          ref_count(1),
+          use_count(0),
+          thunk_handle(thunk_handle),
+          alloc_flag(alloc_flag) {}
+
+    static __forceinline hsa_amd_vmem_alloc_handle_t Convert(void* handle) {
+      hsa_amd_vmem_alloc_handle_t ret_handle = {
+          static_cast<uint64_t>(reinterpret_cast<uintptr_t>(handle))};
+      return ret_handle;
+    }
+
+    __forceinline core::Agent* agentOwner() const { return region->owner(); }
+
+    const MemoryRegion* region;
+    size_t size;
+    int ref_count;
+    int use_count;
+    ThunkHandle thunk_handle;  // handle returned by hsaKmtAllocMemory(NoAddress = 1)
+    MemoryRegion::AllocateFlags alloc_flag;
+  };
+  std::map<ThunkHandle, MemoryHandle> memory_handle_map_;
+
+  struct MappedHandle;
+  struct MappedHandleAllowedAgent {
+    MappedHandleAllowedAgent()
+        : va(NULL), permissions(HSA_ACCESS_PERMISSION_NONE), mappedHandle(NULL), ldrm_bo(0) {}
+    MappedHandleAllowedAgent(MappedHandle* _mappedHandle, Agent* targetAgent, void* va, size_t size,
+                             hsa_access_permission_t perms)
+        : va(va),
+          size(size),
+          targetAgent(targetAgent),
+          permissions(perms),
+          mappedHandle(_mappedHandle),
+          ldrm_bo(0) {}
+
+    hsa_status_t RemoveAccess();
+    hsa_status_t EnableAccess(hsa_access_permission_t perms);
+
+    void* va;
+    size_t size;
+    Agent* targetAgent;
+    hsa_access_permission_t permissions;
+    MappedHandle* mappedHandle;
+    amdgpu_bo_handle ldrm_bo;
+  };
+
+  struct MappedHandle {
+    MappedHandle()
+        : mem_handle(NULL),
+          address_handle(NULL),
+          offset(0),
+          mmap_offset(0),
+          size(0),
+          drm_fd(-1),
+          drm_cpu_addr(NULL),
+          ldrm_bo(0) {}
+
+    MappedHandle(MemoryHandle* mem_handle, AddressHandle* address_handle, uint64_t offset,
+                 size_t size, int drm_fd, void* drm_cpu_addr, hsa_access_permission_t perm,
+                 amdgpu_bo_handle bo)
+        : mem_handle(mem_handle),
+          address_handle(address_handle),
+          offset(offset),
+          mmap_offset(0),
+          size(size),
+          drm_fd(drm_fd),
+          drm_cpu_addr(drm_cpu_addr),
+          ldrm_bo(bo) {}
+
+    __forceinline core::Agent* agentOwner() const { return mem_handle->region->owner(); }
+
+    MemoryHandle* mem_handle;
+    AddressHandle* address_handle;
+    uint64_t offset;
+    uint64_t mmap_offset;
+    size_t size;
+    int drm_fd;
+    void* drm_cpu_addr;  // CPU Buffer address
+    amdgpu_bo_handle ldrm_bo;
+    std::map<Agent*, MappedHandleAllowedAgent> allowed_agents;
+  };
+  std::map<const void*, MappedHandle> mapped_handle_map_;  // Indexed by VA
+
   // Frees runtime memory when the runtime library is unloaded if safe to do so.
   // Failure to release the runtime indicates an incorrect application but is
   // common (example: calls library routines at process exit).
