@@ -70,8 +70,6 @@
 
 namespace rocr {
 namespace AMD {
-// Queue::amd_queue_ is cache-aligned for performance.
-const uint32_t kAmdQueueAlignBytes = 0x40;
 
 HsaEvent* AqlQueue::queue_event_ = nullptr;
 std::atomic<uint32_t> AqlQueue::queue_count_(0);
@@ -80,7 +78,7 @@ int AqlQueue::rtti_id_ = 0;
 
 AqlQueue::AqlQueue(GpuAgent* agent, size_t req_size_pkts, HSAuint32 node_id, ScratchInfo& scratch,
                    core::HsaEventCallback callback, void* err_data, bool is_kv)
-    : Queue(),
+    : Queue(agent->isMES() ? MemoryRegion::AllocateNonPaged : 0),
       LocalSignal(0, false),
       DoorbellSignal(signal()),
       ring_buf_(nullptr),
@@ -285,6 +283,9 @@ AqlQueue::AqlQueue(GpuAgent* agent, size_t req_size_pkts, HSAuint32 node_id, Scr
   queue_id_ = queue_rsrc.QueueId;
   MAKE_NAMED_SCOPE_GUARD(QueueGuard, [&]() { hsaKmtDestroyQueue(queue_id_); });
 
+  // On the first queue creation, reserve some scratch memory on this agent.
+  agent_->ReserveScratch();
+
   // Initialize scratch memory related entities
   queue_scratch_.queue_retry = amd_queue_.queue_inactive_signal;
   InitScratchSRD();
@@ -826,8 +827,12 @@ bool AqlQueue::DynamicScratchHandler(hsa_signal_value_t error_code, void* arg) {
       core::AqlPacket& pkt =
           ((core::AqlPacket*)queue->amd_queue_.hsa_queue.base_address)[pkt_slot_idx];
 
-      assert(pkt.IsValid() && "Invalid packet in dynamic scratch handler.");
-      assert(pkt.type() == HSA_PACKET_TYPE_KERNEL_DISPATCH &&
+      // Load the packet header as atomic acquire as it it written by another
+      // thread as atomic release. This ensures the rest of the packet fields
+      // are visible.
+      uint16_t pkt_header = atomic::Load(&pkt.packet.header, std::memory_order_acquire);
+      assert(core::AqlPacket::IsValid(pkt_header) && "Invalid packet in dynamic scratch handler.");
+      assert(core::AqlPacket::type(pkt_header) == HSA_PACKET_TYPE_KERNEL_DISPATCH &&
              "Invalid packet in dynamic scratch handler.");
       assert((pkt.dispatch.workgroup_size_x != 0) && (pkt.dispatch.workgroup_size_y != 0) &&
              (pkt.dispatch.workgroup_size_z != 0) && "Invalid dispatch dimension.");
@@ -941,7 +946,7 @@ bool AqlQueue::DynamicScratchHandler(hsa_signal_value_t error_code, void* arg) {
         errorCode = HSA_STATUS_ERROR_INVALID_ARGUMENT;
 
       } else if ((error_code & 128) == 128) {  // Out of VGPRs
-        errorCode = HSA_STATUS_ERROR_INVALID_ISA;
+        errorCode = hsa_status_t(HSA_STATUS_ERROR_OUT_OF_REGISTERS);
 
       } else if ((error_code & 0x20000000) == 0x20000000) {  // Memory violation (>48-bit)
         errorCode = hsa_status_t(HSA_STATUS_ERROR_MEMORY_APERTURE_VIOLATION);
@@ -999,43 +1004,43 @@ bool AqlQueue::ExceptionHandler(hsa_signal_value_t error_code, void* arg) {
   };
   static const queue_error_t QueueErrors[] = {
       // EC_QUEUE_WAVE_ABORT
-      1, HSA_STATUS_ERROR_EXCEPTION,
+      { 1, HSA_STATUS_ERROR_EXCEPTION },
       // EC_QUEUE_WAVE_TRAP
-      2, HSA_STATUS_ERROR_EXCEPTION,
+      { 2, HSA_STATUS_ERROR_EXCEPTION },
       // EC_QUEUE_WAVE_MATH_ERROR
-      3, HSA_STATUS_ERROR_EXCEPTION,
+      { 3, HSA_STATUS_ERROR_EXCEPTION },
       // EC_QUEUE_WAVE_ILLEGAL_INSTRUCTION
-      4, (hsa_status_t)HSA_STATUS_ERROR_ILLEGAL_INSTRUCTION,
+      { 4, (hsa_status_t)HSA_STATUS_ERROR_ILLEGAL_INSTRUCTION },
       // EC_QUEUE_WAVE_MEMORY_VIOLATION
-      5, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_FAULT,
+      { 5, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_FAULT },
       // EC_QUEUE_WAVE_APERTURE_VIOLATION
-      6, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_APERTURE_VIOLATION,
+      { 6, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_APERTURE_VIOLATION },
       // EC_QUEUE_PACKET_DISPATCH_DIM_INVALID
-      16, HSA_STATUS_ERROR_INCOMPATIBLE_ARGUMENTS,
+      { 16, HSA_STATUS_ERROR_INCOMPATIBLE_ARGUMENTS },
       // EC_QUEUE_PACKET_DISPATCH_GROUP_SEGMENT_SIZE_INVALID
-      17, HSA_STATUS_ERROR_INVALID_ALLOCATION,
+      { 17, HSA_STATUS_ERROR_INVALID_ALLOCATION },
       // EC_QUEUE_PACKET_DISPATCH_CODE_INVALID
-      18, HSA_STATUS_ERROR_INVALID_CODE_OBJECT,
+      { 18, HSA_STATUS_ERROR_INVALID_CODE_OBJECT },
       // EC_QUEUE_PACKET_UNSUPPORTED
-      20, HSA_STATUS_ERROR_INVALID_PACKET_FORMAT,
+      { 20, HSA_STATUS_ERROR_INVALID_PACKET_FORMAT },
       // EC_QUEUE_PACKET_DISPATCH_WORK_GROUP_SIZE_INVALID
-      21, HSA_STATUS_ERROR_INVALID_ARGUMENT,
+      { 21, HSA_STATUS_ERROR_INVALID_ARGUMENT },
       // EC_QUEUE_PACKET_DISPATCH_REGISTER_SIZE_INVALID
-      22, HSA_STATUS_ERROR_INVALID_ISA,
+      { 22, HSA_STATUS_ERROR_INVALID_ISA },
       // EC_QUEUE_PACKET_VENDOR_UNSUPPORTED
-      23, HSA_STATUS_ERROR_INVALID_PACKET_FORMAT,
+      { 23, HSA_STATUS_ERROR_INVALID_PACKET_FORMAT },
       // EC_QUEUE_PREEMPTION_ERROR
-      31, HSA_STATUS_ERROR,
+      { 31, HSA_STATUS_ERROR },
       // EC_DEVICE_MEMORY_VIOLATION
-      33, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_APERTURE_VIOLATION,
+      { 33, (hsa_status_t)HSA_STATUS_ERROR_MEMORY_APERTURE_VIOLATION },
       // EC_DEVICE_RAS_ERROR
-      34, HSA_STATUS_ERROR,
+      { 34, HSA_STATUS_ERROR },
       // EC_DEVICE_FATAL_HALT
-      35, HSA_STATUS_ERROR,
+      { 35, HSA_STATUS_ERROR },
       // EC_DEVICE_NEW
-      36, HSA_STATUS_ERROR,
+      { 36, HSA_STATUS_ERROR },
       // EC_PROCESS_DEVICE_REMOVE
-      50, HSA_STATUS_ERROR};
+      { 50, HSA_STATUS_ERROR }};
 
   AqlQueue* queue = (AqlQueue*)arg;
   hsa_status_t errorCode = HSA_STATUS_ERROR;
@@ -1059,7 +1064,7 @@ bool AqlQueue::ExceptionHandler(hsa_signal_value_t error_code, void* arg) {
 
   // Suppress VM fault reporting.  This is more useful when reported through the system error
   // handler.
-  if (errorCode == HSA_STATUS_ERROR_MEMORY_FAULT) {
+  if (errorCode == static_cast<hsa_status_t>(HSA_STATUS_ERROR_MEMORY_FAULT)) {
     debug_print("Queue error - HSA_STATUS_ERROR_MEMORY_FAULT\n");
     return false;
   }
@@ -1191,6 +1196,8 @@ void AqlQueue::ExecutePM4(uint32_t* cmd_data, size_t cmd_size_b) {
   // To respect multi-producer semantics, first buffer commands for the queue slot.
   constexpr uint32_t slot_size_dw = uint32_t(slot_size_b / sizeof(uint32_t));
   uint32_t slot_data[slot_size_dw];
+  hsa_signal_t signal = {0};
+  hsa_status_t err;
 
   if (agent_->isa()->GetMajorVersion() <= 8) {
     // Construct a set of PM4 to fit inside the AQL packet slot.
@@ -1234,23 +1241,28 @@ void AqlQueue::ExecutePM4(uint32_t* cmd_data, size_t cmd_size_b) {
     // Construct an AQL packet to jump to the PM4 IB.
     struct amd_aql_pm4_ib {
       uint16_t header;
-      uint16_t ven_hdr;
+      uint8_t  amd_format;
+      uint8_t  reserved0;
       uint32_t ib_jump_cmd[4];
       uint32_t dw_cnt_remain;
-      uint32_t reserved[8];
+      uint32_t reserved1[8];
       hsa_signal_t completion_signal;
     };
 
     constexpr uint32_t AMD_AQL_FORMAT_PM4_IB = 0x1;
 
+    err = hsa_signal_create(1, 0, NULL, &signal);
+    assert(err == HSA_STATUS_SUCCESS);
+
     amd_aql_pm4_ib aql_pm4_ib{};
     aql_pm4_ib.header = HSA_PACKET_TYPE_VENDOR_SPECIFIC << HSA_PACKET_HEADER_TYPE;
-    aql_pm4_ib.ven_hdr = AMD_AQL_FORMAT_PM4_IB;
+    aql_pm4_ib.amd_format = AMD_AQL_FORMAT_PM4_IB;
     aql_pm4_ib.ib_jump_cmd[0] = ib_jump_cmd[0];
     aql_pm4_ib.ib_jump_cmd[1] = ib_jump_cmd[1];
     aql_pm4_ib.ib_jump_cmd[2] = ib_jump_cmd[2];
     aql_pm4_ib.ib_jump_cmd[3] = ib_jump_cmd[3];
     aql_pm4_ib.dw_cnt_remain = 0xA;
+    aql_pm4_ib.completion_signal = signal;
 
     memcpy(slot_data, &aql_pm4_ib, sizeof(aql_pm4_ib));
   } else {
@@ -1268,10 +1280,15 @@ void AqlQueue::ExecutePM4(uint32_t* cmd_data, size_t cmd_size_b) {
   doorbell->StoreRelease(write_idx);
 
   // Wait for the packet to be consumed.
-  // Should be switched to a signal wait when aql_pm4_ib can be used on all
-  // supported platforms.
-  while (queue->LoadReadIndexRelaxed() <= write_idx) {
-    os::YieldThread();
+  if (agent_->isa()->GetMajorVersion() <= 8) {
+    while (queue->LoadReadIndexRelaxed() <= write_idx)
+      os::YieldThread();
+  } else {
+    hsa_signal_value_t ret;
+    ret = hsa_signal_wait_scacquire(signal, HSA_SIGNAL_CONDITION_LT, 1,
+                                    (uint64_t)-1, HSA_WAIT_STATE_ACTIVE);
+    err = hsa_signal_destroy(signal);
+    assert(ret == 0 && err == HSA_STATUS_SUCCESS);
   }
 }
 
@@ -1318,7 +1335,11 @@ void AqlQueue::FillBufRsrcWord1_Gfx11() {
 
 void AqlQueue::FillBufRsrcWord2() {
   SQ_BUF_RSRC_WORD2 srd2;
-  srd2.bits.NUM_RECORDS = uint32_t(queue_scratch_.size);
+  const auto& agent_props = agent_->properties();
+  const uint32_t num_xcc = agent_props.NumXcc;
+
+   // report size per XCC
+  srd2.bits.NUM_RECORDS = uint32_t(queue_scratch_.size / num_xcc);
 
   amd_queue_.scratch_resource_descriptor[2] = srd2.u32All;
 }
@@ -1389,8 +1410,10 @@ void AqlQueue::FillComputeTmpRingSize() {
     return;
   }
 
-  // Determine the maximum number of waves device can support
   const auto& agent_props = agent_->properties();
+  const uint32_t num_xcc = agent_props.NumXcc;
+
+  // Determine the maximum number of waves device can support
   uint32_t num_cus = agent_props.NumFComputeCores / agent_props.NumSIMDPerCU;
   uint32_t max_scratch_waves = num_cus * agent_props.MaxSlotsScratchCU;
 
@@ -1402,10 +1425,11 @@ void AqlQueue::FillComputeTmpRingSize() {
   tmpring_size.bits.WAVESIZE = wave_scratch;
   assert(wave_scratch == tmpring_size.bits.WAVESIZE && "WAVESIZE Overflow.");
   uint32_t num_waves =
-      queue_scratch_.size / (tmpring_size.bits.WAVESIZE * queue_scratch_.mem_alignment_size);
+      (queue_scratch_.size / num_xcc) / (tmpring_size.bits.WAVESIZE * queue_scratch_.mem_alignment_size);
+
   tmpring_size.bits.WAVES = std::min(num_waves, max_scratch_waves);
   amd_queue_.compute_tmpring_size = tmpring_size.u32All;
-  assert((tmpring_size.bits.WAVES % agent_props.NumShaderBanks == 0) &&
+  assert((tmpring_size.bits.WAVES % (agent_props.NumShaderBanks / num_xcc) == 0) &&
          "Invalid scratch wave count.  Must be divisible by #SEs.");
 }
 
@@ -1417,9 +1441,11 @@ void AqlQueue::FillComputeTmpRingSize_Gfx11() {
     return;
   }
 
-  // Determine the maximum number of waves device can support
   const auto& agent_props = agent_->properties();
-  uint32_t num_cus = agent_props.NumFComputeCores / agent_props.NumSIMDPerCU;
+  const uint32_t num_xcc = agent_props.NumXcc;
+
+  // Determine the maximum number of waves device can support
+  uint32_t num_cus = agent_props.NumFComputeCores / (agent_props.NumSIMDPerCU * num_xcc);
   uint32_t max_scratch_waves = num_cus * agent_props.MaxSlotsScratchCU;
 
   // Scratch is allocated program COMPUTE_TMPRING_SIZE register
@@ -1469,7 +1495,11 @@ void AqlQueue::InitScratchSRD() {
 
   // Populate flat scratch parameters in amd_queue_.
   amd_queue_.scratch_backing_memory_location = queue_scratch_.queue_process_offset;
-  amd_queue_.scratch_backing_memory_byte_size = queue_scratch_.size;
+
+  const auto& agent_props = agent_->properties();
+  const uint32_t num_xcc = agent_props.NumXcc;
+  // report size per XCC
+  amd_queue_.scratch_backing_memory_byte_size = queue_scratch_.size / num_xcc;
 
   // For backwards compatibility this field records the per-lane scratch
   // for a 64 lane wavefront. If scratch was allocated for 32 lane waves
